# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.


import pytest
from types import SimpleNamespace
import time
import jsonschema
from jsonschema import FormatChecker
from pyapix.tool.tools import *


class info:
  class local:
    class swagger:
        petstore = '~/local/petstore/swagger.json'
        nws = '~/local/nws/openapi.json'
        worms =  '~/local/worms/openapi.yaml'
  class remote:
    class swagger:
        petstore = 'https://petstore.swagger.io/v2/swagger.json'
        nws = 'https://api.weather.gov/openapi.json'


def test_dvalidator(): 

    local_validate = lambda params: None
    schema = {'required': ['ScientificName'], 'properties': {'ScientificName': {'type': 'string'}, 'like': {'type': 'boolean', 'default': 'true'}, 'marine_only': {'type': 'boolean', 'default': 'true'}, 'offset': {'type': 'integer', 'default': 1}}, 'additionalProperties': False, 'type': 'object'}
    validator = dvalidator(local_validate)(schema, format_checker=FormatChecker())

    params = {'ScientificName': 'Solea solea'}
    assert validator.is_valid(params) is True
    assert validator.validate(params) is None

    params = {'taxonid': 127160}
    assert validator.is_valid(params) is False
    with pytest.raises(jsonschema.exceptions.ValidationError):
        validator.validate(params)

    params = {}
    assert validator.is_valid(params) is False
    with pytest.raises(jsonschema.exceptions.ValidationError):
        validator.validate(params)


def test_retry():

    @retry_call()
    def fun(url, verb, request_params):
        return SimpleNamespace(is_success=True)

    tr = fun(0, 0, 0)
    assert tr.is_success is True

    @retry_call(n=2)
    def fun(url, verb, request_params):
        return SimpleNamespace(is_success=False)

    t0 = time.time()
    tr = fun(0, 0, 0)
    td = time.time() - t0
    assert tr.is_success is False
    assert 3 < td < 3.1


def test_loading():
    foo = dict(local=info.local, remote=info.remote)
    for key in foo:
        print(key)
        thing = foo[key]
        dct = dict(thing.swagger.__dict__)
        keys = [k for k in dct if not k.startswith('__')]
        for key in keys:
            jdoc = parsed_file_or_url(dct[key])
            assert type(jdoc) is dict
            assert len(jdoc) > 3
            print(key, len(jdoc))
        print()


def test_extract_from_dict_list():
    dlist = [
        dict(name='first', a=1, b=2),
        dict(name='second', a=3, b=4),
        dict(name='third', a=5, b=6, c=7),
    ]
    foo = extract_from_dict_list(dlist, 'a')
    assert foo == {'first': 1, 'second': 3, 'third': 5}
    
    foo = extract_from_dict_list(dlist, 'b')
    assert foo == {'first': 2, 'second': 4, 'third': 6}
    
    foo = extract_from_dict_list(dlist, 'c')
    assert foo == {'third': 7}
    
    foo = extract_from_dict_list(dlist, 'd')
    assert foo == {}
    
    foo = extract_from_dict_list({}, 'd')
    assert foo == {}

    d = dict(e=8)
    with pytest.raises(TypeError):
        extract_from_dict_list(d, 'e')
    
    with pytest.raises(KeyError):
        extract_from_dict_list([d], 'e')
 

def test_raw_file_or_url():
    pet_local = raw_file_or_url(info.local.swagger.petstore)
    pet_remote = raw_file_or_url(info.remote.swagger.petstore)
    ll = len(pet_local)
    lr = len(pet_remote)

    nws_local = raw_file_or_url(info.local.swagger.nws)
    nws_remote = raw_file_or_url(info.remote.swagger.nws)
    ll = len(nws_local)
    lr = len(nws_remote)


def test_parsed_file_or_url():
    pet_local = parsed_file_or_url(info.local.swagger.petstore)
    pet_remote = parsed_file_or_url(info.remote.swagger.petstore)
    assert list(pet_local) == list(pet_remote)

    nws_local = parsed_file_or_url(info.local.swagger.nws)
    nws_remote = parsed_file_or_url(info.remote.swagger.nws)
    assert list(nws_local) == list(nws_remote)
    assert list(nws_local['paths']) == list(nws_remote['paths'])


def xtest_toml():
    import tomllib    # requires 3.11
    tpath = 'foo.toml'
    with open(tpath) as fh:
        ts = fh.read()
    tdoc = tomllib.loads(ts)
    # TODO: note, this is more of a demo than a test.
    # Prep for using toml.




# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from datetime import datetime

from .info import local
from pyapix.tool.tools import LocalValidationError, ValidDataBadResponse
from pyapix.tool.api_tools import dynamic_validator, dynamic_call, SurpriseArgs, Service
from pyapix.tool import api_tools
from pyapix.tool.api_tools import *


class Foo(LocalValidationError): pass


def local_validate(params):
    """Catch data problems missed by the schema.
    # eg start_date > end_date
    params = {
        'start': '2024-09-17T18:39:00+00:00', 
        'end':   '2024-09-18T18:39:00+00:00',
    }
    """
    fmt = '%Y-%m-%dT%H:%M:%S+00:00'


def param_func(pdict):
  try:
    """
    Extract info for inserting into schema.
    """
    d = {}
#    for key in ['required', 'name', 'schema', 'type']:
    for key in ['required', 'name', 'schema']:
        if key in pdict:
            d[key] = pdict[key]
    if 'schema' not in d:
        d['schema'] = {}
        for key in ['format', 'enum', 'type', 'items', 'default']:
            if key in pdict:
                d['schema'][key] = pdict[key]

        # below is required because otherwise we get a jsonschema error...
        # jsonschema.exceptions.UnknownType: Unknown type 'file'
        if d['schema']['type'] == 'file': 
            d['schema']['type'] = 'string'

    return d
  finally:
    pass


def altered_raw_swagger(jdoc):
  try:
    """Alter raw data to conform with local code assumptions.
    This function takes a swagger doc as a json and returns json.
    """
    for endpoint in jdoc['paths']:
        epdoc = jdoc['paths'][endpoint]
        for verb in epdoc:
            vdoc = epdoc[verb]
            parameters = vdoc['parameters']
            jdoc['paths'][endpoint][verb]['parameters'] = [param_func(d) for d in parameters]
            if (endpoint, verb) == ('/pet/findByStatus', 'get'):
                foop = parameters
    return jdoc
  finally:
    pass



def prep_func(config):
    jdoc = jsonref.loads(json.dumps(parsed_file_or_url(config.swagger_path)))
    paths = config.alt_swagger(jdoc)['paths']
    head_func = config.head_func

    def prepped(endpoint, verb, args):
      try:
        """
        """
        if not args:
            return (config.api_base + endpoint, verb, {})
        if type(args) not in (dict, list, str):
            raise Exception(f'{endpoint} {verb} {args}')
        if type(args) is not dict:
            raise NonDictArgs(f'{endpoint} {verb} {args}')

        ev_params = paths[endpoint][verb]['parameters'] or {}
        location = extract_from_dict_list(ev_params, 'in')
#         print('prepped')
#         print(endpoint, verb, list(ev_params))
#         print('location', location)
#         print('......')
        request_params = {}
        query = {}
        form_data = {}
        files = {}
        data = {}
        headers = {}

        assert type(args) is dict
#         print()
#         print(f'xx ep: {endpoint}  verb: {verb}')
        for arg in args:

            # TODO: 'query' is not a good default.
            # TODO: 'query' is not a good default.
            plocation = location[arg] if arg in location else 'query'
            # TODO: 'query' is not a good default.
            # TODO: 'query' is not a good default.
            # TODO: change altered_raw_swagger to always have location.
            # because there should not be a default anyway.
            if '{'+str(arg)+'}' in endpoint:
                plocation = 'path'

            # TMP    #  petstore-specific    #  petstore-specific
            if arg == 'file':                  #  petstore-specific
                plocation = 'files'         #  petstore-specific
            if arg == 'additionalMetadata':    #  petstore-specific
                plocation = 'data'             #  petstore-specific
                # TODO: find another API that does file uploads.
            # TMP    #  petstore-specific    #  petstore-specific

            #  petstore-specific
            if endpoint.startswith('/pet') and verb == 'post': 
                sp = endpoint.split('/')
                if len(sp) == 3:
                    try:
                        int(sp[-1])
#                         print('yoohoo')
                        plocation = 'data'             #  petstore-specific
                    except ValueError:
                        pass
#                 print(endpoint, sp, arg)
                
#             print(f'xx arg: {arg}  plocation: {plocation}')

            if plocation == 'path':
                endpoint = endpoint.replace('{'+arg+'}', str(args[arg]))
            elif plocation == 'query':
                query[arg] = args[arg]
            elif plocation == 'header':
                headers[arg] = args[arg]
            elif plocation == 'formData':
                form_data[arg] = args[arg]
            elif plocation == 'files':
                files[arg] = args[arg]
            elif plocation == 'data':
                data[arg] = args[arg]

#         print()
        if query:
            request_params['params'] = query
        if verb in 'post put' and 'params' in request_params:
            request_params = {'json': request_params['params']}
        if headers:
            request_params['headers'] = headers
        if data:
            request_params['data'] = data
        if files:
            request_params['files'] = files

        heads = head_func(endpoint, verb) if head_func else {}
        if heads:
            if not 'headers' in request_params:
                request_params['headers'] = {}
            request_params['headers'].update(heads)

        return (config.api_base + endpoint, verb, request_params)
      finally:
        pass
    return prepped


def parameters_to_schema(ev_info):
  try:
    if 'parameters' in ev_info:
        parameters = ev_info['parameters'] or {}
    else:
        parameters = {}

    # pattern matching makes the case logic MUCH simpler.
    match parameters:
        # TODO: note, both of these cases are for petstore.
        # So maybe move the logic to the petshop client.
        # BUT otoh, the func is indeed imperfect.
        # 'type': 'object',     # NOOOOOOOOOOOOOOOOOOOO!!!!
        # fails when parameters is a list instead of a dict.
        # thus
        # # TODO: figure out what the real problem is.
        # and fix in the appropriate place.
        case [{'name': thing}] if thing=='body':
#             print('.....................................yoohoo body GUARD')
            return parameters[0]['schema']
        case [{'name': 'body'}]:   # identical to above and therefore unreachable.
#             print('.....................................yoohoo body')
            return parameters[0]['schema']
        case [{'name': 'status'}]:  # {'name': 'status'} is definitely petstore-specific
#             print('yoohoo status')
            return parameters[0]

    pr = extract_from_dict_list(parameters, 'required')
    return {
        'required': [key for key in pr if pr[key]],
        'properties': extract_from_dict_list(parameters, 'schema'), 
        'additionalProperties': False, 
        'type': 'object',     # OK
    }
  finally:
    pass

        
def head_func(endpoint, verb):
    return {'user-agent': 'python-httpx/0.27.2'}


class config:
    swagger_path = local.swagger.petstore
    api_base = local.api_base.petstore
    alt_swagger = altered_raw_swagger
    head_func = head_func
    validate = local_validate


api_tools.prep_func = prep_func
api_tools.parameters_to_schema = parameters_to_schema


_validator = dynamic_validator(config)
call = dynamic_call(config)


from pyapix.tool.tools import parsed_file_or_url
from pyapix.tool.api_tools import endpoints_and_verbs
jdoc = parsed_file_or_url(config.swagger_path)
ends = endpoints_and_verbs(jdoc)
service = Service('Petstore', call, _validator, ends)





# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

import json
import jsonref
from jsonschema import FormatChecker
import httpx
from pyapix.tool.tools import (dvalidator, parsed_file_or_url, identity_func,
                   extract_from_dict_list,
                   retry_call,
                   LocalValidationError,
                   raw_file_or_url,
                   )

class ValidDataBadResponse(LocalValidationError): pass

class SurpriseArgs(Exception): pass
class NonDictArgs(SurpriseArgs): pass
class ListArgs(SurpriseArgs): pass

class BadEndpointOrVerb(Exception): pass


def endpoints_and_verbs(jdoc):
    return [(p,v) for p in jdoc['paths'] for v in jdoc['paths'][p]]


def parameters_to_schema(ev_info):
    if 'parameters' in ev_info:
        parameters = ev_info['parameters']
    pr = extract_from_dict_list(parameters or {}, 'required')
    return {
        'required': [key for key in pr if pr[key]],
        'properties': extract_from_dict_list(parameters or {}, 'schema'), 
        'additionalProperties': False, 
        'type': 'object',
    }


def dynamic_validator(config):
    local_validate = config.validate
    jdoc = jsonref.loads(json.dumps(parsed_file_or_url(config.swagger_path)))
    paths = config.alt_swagger(jdoc)['paths']

    def validator(endpoint, verb='get'):
        """Return a validator for `(endpoint, verb)`.
        """
        try:
            schema = parameters_to_schema(paths[endpoint][verb])
        except KeyError as exc:
            raise BadEndpointOrVerb(exc.args[0])
        return dvalidator(local_validate)(schema, format_checker=FormatChecker())
    return validator


def dynamic_call(config):
    prepped = prep_func(config)

    @retry_call()
    def call(endpoint, verb, params):
        """Call (endpoint, verb) with params. """
        (url, verb, request_params) = prepped(endpoint, verb, params)
        request = httpx.Request(verb, url, **request_params)
        with httpx.Client() as client:
            return client.send(request)  
    return call


def prep_func(config):
    jdoc = jsonref.loads(json.dumps(parsed_file_or_url(config.swagger_path)))
    paths = config.alt_swagger(jdoc)['paths']
    head_func = config.head_func

    def prepped(endpoint, verb, args):
      try:
        """Prepare args for endpoint, verb.
        """
        if not args:
            return (config.api_base + endpoint, verb, {})
        if type(args) is not dict:
            raise NonDictArgs(f'{endpoint} {verb} {args}')

        ev_params = paths[endpoint][verb]['parameters'] or {}
        location = extract_from_dict_list(ev_params, 'in')
        request_params = {}
        query = {}
        form_data = {}
        files = {}
        data = {}
        headers = {}

        assert type(args) is dict
        for arg in args:
            plocation = location[arg]
            if '{'+str(arg)+'}' in endpoint:
                plocation = 'path'

            if plocation == 'path':
                endpoint = endpoint.replace('{'+arg+'}', str(args[arg]))
            elif plocation == 'query':
                query[arg] = args[arg]
            elif plocation == 'header':
                headers[arg] = args[arg]
            elif plocation == 'formData':
                form_data[arg] = args[arg]
            elif plocation == 'files':
                files[arg] = args[arg]
            elif plocation == 'data':
                data[arg] = args[arg]

        if query:
            request_params['params'] = query
        if verb in 'post put' and 'params' in request_params:
            request_params = {'json': request_params['params']}
        if headers:
            request_params['headers'] = headers
        if data:
            request_params['data'] = data
        if files:
            request_params['files'] = files

        heads = head_func(endpoint, verb) if head_func else {}
        if heads:
            if not 'headers' in request_params:
                request_params['headers'] = {}
            request_params['headers'].update(heads)

        return (config.api_base + endpoint, verb, request_params)
      finally:
        pass
    return prepped


class Service:
    def __init__(self, name, call, _validator, ends):
        self.name = name
        self.call = call
        self._validator = _validator
        self.ends = ends

    def __repr__(self):
        return f"Service('{self.name}')"





# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from datetime import datetime

from pyapix.tool.tools import ( LocalValidationError,)
from pyapix.tool.api_tools import (dynamic_validator, dynamic_call,)
from .info import local

class DateOrderError(LocalValidationError): pass

class ValidDataBadResponse(LocalValidationError): pass


def local_validate(params):
    """Catch data problems missed by the schema.
    # eg start_date > end_date
    params = {
        'start': '2024-09-17T18:39:00+00:00', 
        'end':   '2024-09-18T18:39:00+00:00',
    }
    """
    fmt = '%Y-%m-%dT%H:%M:%S+00:00'
    if 'start' in params and 'end' in params:
        start = params['start']
        end = params['end']
        if datetime.strptime(start, fmt) > datetime.strptime(end, fmt): 
            raise DateOrderError(start, end)


def altered_raw_swagger(jdoc):
    """Alter raw data to conform with local code assumptions.
    This function takes a swagger doc as a json and returns json.
    """
    x = jdoc['components']['parameters']['datasetid']
    jdoc['components']['parameters']['id_dataset'] = x
    return jdoc


class config:
    swagger_path = local.swagger.obis
    api_base = local.api_base.obis
    alt_swagger = altered_raw_swagger
    head_func = lambda endpoint, verb: {}
    validate = local_validate


_validator = dynamic_validator(config)
call = dynamic_call(config)





import os
import copy
import json
from jsonschema import Draft7Validator, FormatChecker

from pyapix.tool.exploratory import pop_key
from pyapix.tool.api_tools import BadEndpointOrVerb, Service, endpoints_ands_verbs
from pyapix.tool.tools import parsed_file_or_url, list_of_dict_to_dict

from working_with_postman import (is_request, show_contents, insert_params, 
          fetch_thing, is_bad_schema, make_endpoint, )

import crs_conversion_api as con
import crs_catalog_api as cat
import entitlements_api as entitlements
import legal_api as legal
import unit_api as unit


one_dict = list_of_dict_to_dict()

service_parts = ['api', 'crs', 'catalog', 'conversion', 'search', 'storage',
                 'legal'
]

# iterate over Postman file.                   DONE
# Find all calls to CRS Conversion.            DONE
# For each call, 
#     translate url(PM) to endpoint(pyapiX)    DONE
#     grab the parameters                      DONE
#     validate the parameters                  WIP
# create client for crs/catalog                DONE
#     validate the parameters                  DONE
# store the parameters in yaml? (or PM?)


def pm_files():
    preship_dir = '~/osdu/pre-shipping/R3-M24/AWS-M24'
    paths = []
    dname = 'Core Services'
    fnames = [
        'AWS_OSDUR3M24_CoreServices_Collection.postman_collection.json',
#        'AWS_OSDUR3M24_VersionEndPoints_Collection.postman_collection.json'
    ]
    for fname in fnames:
        fpath = os.path.expanduser(f'{preship_dir}/{dname}/{fname}')
        paths.append(fpath)
#     dname = 'Policy'
#     fname = 'AWS_OSDUR3M24_Policy_Collection.postman_collection.json'
#     fpath = os.path.expanduser(f'{preship_dir}/{dname}/{fname}')
#     paths.append(fpath)
    return paths


# (almost) nothing specific to CRS here.
def request_for_service(service):
    def do_pm_request(postman_request):
      try:
        pr = postman_request['request']
        if 'body' in pr:
            ctrb = pr['body']
    #        assert sorted(list(ctrb)) == ['mode', 'options', 'raw']
            if (ctrb['mode'] == 'raw') and ctrb['raw']:
                bdecoded = json.loads(ctrb['raw'])
            else:
                bdecoded = ''
        else:
            bdecoded = ''

        url = pr['url']
        params = one_dict(url['query']) if 'query' in url else {}
        if bdecoded:
            params['body'] = bdecoded

        endpoint = make_endpoint(pr['url'], service)
        verb = pr['method'].lower()

#         entitlement_params = dict(
#             group_email = 'xxxxx',
#         )
#         params.update(entitlement_params)
#         # TODO: insert path vars

        # TODO: this stuff is hard-coded.
        # Source will be an Environment.
        # Targets will be anything in parameters or headers.
        dpid = one_dict(pr['header'])['data-partition-id']
        source = dict(data_partition_id='foo..dpi..bar')
        subbed = insert_params(dpid, source)
        # TODO: strategy...
        # serialize params
        # substitute
        # deserialize params

        bep = False
        vn = None
        try:
            v = service._validator(endpoint, verb)
        except BadEndpointOrVerb:
            bep = True
            bad_ones = ['/legaltags/', '/api/unit/actuator']
            assert any(endpoint.startswith(bw) for bw in bad_ones)
        if bep:
            valid_params = 'weird endpoint\n'
            valid_params += '>>>>>>>>>>>> ' + endpoint + '\n'
            valid_params += '>>>>>>>>>>>> ' + endpoint + '\n'
        else:
            try:
                # TODO: this is for debugging Unit service validation.
                # It seems many of the endpoints have schemas that are not quite
                # correct.
                # TODO: repair the swagger doc.   ?????  maybe
                # It's a bit of dilemma because the schema contains information
                # that is likely useful but is not quite relevant to the
                # parameters.
                vn = Draft7Validator(v.v.schema['properties']['request'], format_checker=FormatChecker())
            except KeyError:
                pass
            schema = v.v.schema   # in case we want to have a look.
            valid_params = 'OK' if v.is_valid(params) else 'invalid params'
            if vn and (valid_params == 'invalid params'):
                valid_params = 'OKxxxx' if vn.is_valid(params) else 'still invalid'
            if is_bad_schema(schema):
                print('yoohoo bad schema')
                print('yoohoo bad schema')
                print('yoohoo bad schema')
                print('yoohoo bad schema')
                cs = schema
                csp = params
                csep = endpoint
                cv = verb
                valid_params = 'crap schema'
                assert endpoint, verb == ('/coordinate-reference-system', 'get')
                # TODO: this (endpoint, verb) has no useful schema.
                # What to do about it?
            # TODO: call the endpoint with the params.

        print(postman_request['name'])
        print(endpoint, verb)
        print('params', params)
        print('............', valid_params)
        print()
      finally:
        globals().update(locals())
    return do_pm_request


def test_pm_section(pmjdoc, *names):
  try:
    """
    Grab arbitrary section from Postman file and do the PM requests.
    """
    postman_item = fetch_thing(pmjdoc, *names)
    rnames = [thing['name'] for thing in postman_item['item']]
    service = smap[names[1]]
    do_pm_request = request_for_service(service)
    for name in rnames:
        if name == 'Health Check':
            print('yoohoo', name)
            print('yoohoo', name)
            return
        noms = names + (name,)
        thing = fetch_thing(pmjdoc, *noms)
        if is_request(thing):
            print('noms', noms)
            rf = do_pm_request(thing)
            if noms[-1] == 'postUnitSystem':
                print('yoohoo xxxxxxxxxxxxx')
                print('yoohoo xxxxxxxxxxxxx')
                foo = copy.deepcopy(locals())
                return                            # tmp for debugging
                break                             # tmp for debugging
                                                  # but strangely, it does not
                                                  # work.
        else:    # it is an `item`.
            print(';;;;;;;;;;;;;;;;;;;;', noms)
            print(';;;;;;;;;;;;;;;;;;;;', 'doing an item... ')
            test_pm_section(pmjdoc, *noms)
  finally:
    globals().update(locals())


# TODO: rename to test_endpoints_ands_verbs    ??????
def test_inspect_swagger():
    """
    CRS Catalog
    Have a look at one of the endpoints from the swagger file.
    """
    jdoc = parsed_file_or_url(config.swagger_path)
    evs = endpoints_ands_verbs(jdoc)[:-1][:1]   # just the first endpoint
    evs = endpoints_ands_verbs(jdoc)[:-1]   # ignore /info
    evs = endpoints_ands_verbs(jdoc)
    for (e, v) in evs:
        print(e, v)
        ev = jdoc['paths'][e][v]
    defs = jdoc['definitions']
    point = defs['Point']
    ps = defs['PointsInAOUSearch']
    # TODO: leverage examples  in definitions.


# Test individual services
# ######################################################################### #

smap = {
    'CRS Catalog': cat.service,
    'CRS Conversion': con.service,
    'Entitlements': entitlements.service,
    'Legal': legal.service,
    'Unit': unit.service,
}


def test_legal():
  try:
    pmjdoc = parsed_file_or_url(pm_files()[0])
    names = ['Core Services', 'Legal',]
    print(names[1])
    test_pm_section(pmjdoc, *names)
    print()
  finally:
    globals().update(locals())


def test_entitlements():
  try:
    pmjdoc = parsed_file_or_url(pm_files()[0])
    names = ['Core Services', 'Entitlements',]
    print(names[1])
    test_pm_section(pmjdoc, *names)
    print()
  finally:
    globals().update(locals())


def test_unit():
  try:
    """Unit service has a complex hierarchy of items & requests.
    """
    pmjdoc = parsed_file_or_url(pm_files()[0])
    names = ['Core Services', 'Unit', 'v3', 'catalog']
    names = ['Core Services', 'Unit', 'v3', 'conversion']
    names = ['Core Services', 'Unit', 'v3', 'measurement']
    names = ['Core Services', 'Unit', 'v3', 'unitsystem']
    names = ['Core Services', 'Unit', 'v3', 'unit', 'unitsystem']
    names = ['Core Services', 'Unit', 'v3', 'unit', 'measurement']
    names = ['Core Services', 'Unit', 'v3', 'unit', 'measurement', 'preferred']
    names = ['Core Services', 'Unit', 'v3', 'unit']    # has both
    names = ['Core Services', 'Unit']    # has both
    print(names[1])
    test_pm_section(pmjdoc, *names)
    print()
  finally:
    globals().update(locals())


def test_crs_catalog():
    pmjdoc = parsed_file_or_url(pm_files()[0])
    names = ['Core Services', 'CRS Catalog', 'V3']
    print(names[1])
    test_pm_section(pmjdoc, *names)
    print()


def test_crs_conversion():
    pmjdoc = parsed_file_or_url(pm_files()[0])
    print()
    names = ['Core Services', 'CRS Conversion', 'v3', 'convert']
    print(names[1])
    test_pm_section(pmjdoc, *names)
    names = ['Core Services', 'CRS Conversion', 'v3', 'convertTrajectory']
    test_pm_section(pmjdoc, *names)
    names = ['Core Services', 'CRS Conversion', 'v3', 'convertGeoJson']
    test_pm_section(pmjdoc, *names)
    names = ['Core Services', 'CRS Conversion', 'V3', 'v3', 'convertBinGrid']
    names = ['Core Services', 'CRS Conversion', 'v3', 'convertBinGrid']
#     test_pm_section(pmjdoc, *names)
# No such endpoint in 'v4'  swagger
    # TODO: note inconsistency with show_contents.
#     show_contents(pmjdoc, 'Core Services', 'CRS Conversion', 'v3', 'convertBinGrid')
# 
# but wait...
# Where did the 'V3' come from?
# It appeared at some point in my explorations.
# But seems to be superfluous.
# TODO: it must be related to the bug in show_contents


def test_all():
    test_crs_catalog()
    test_crs_conversion()
    test_legal()
    test_entitlements()
    test_unit()




from functools import wraps
from inspect import signature


def pop_inputs(fun):
    """
    Remove input args from the return value of a function, if present.
    >>> @pop_inputs
    ... def f(a, x=1):
    ...     y = 2
    ...     return locals()
    ...
    >>> f(0)
    {'y': 2}
    """
    param_names = list(signature(fun).parameters)
    @wraps(fun)
    def inner(*pos, **kw):
        result = fun(*pos, **kw)
        [result.pop(key) for key in param_names if key in result]
        return result
    return inner


def pop_key(name):
    """
    Remove named output arg, if present.
    >>> @pop_key('x')
    ... def f():
    ...     x = 1
    ...     y = 2
    ...     return locals()
    ...
    >>> f()
    {'y': 2}
    """
    def deco(fun):
        @wraps(fun)
        def inner(*arg, **kw):
            result = fun(*arg, **kw)
            if name in result:
                result.pop(name)
            return result
        return inner
    return deco


if __name__ == '__main__':
    import doctest
    doctest.testmod()




# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from collections import defaultdict

from pyapix.tool.api_tools import NonDictArgs
from pyapix.tool.tools import parsed_file_or_url     #, raw_swagger
from pyapix.client.protein import _validator, call, altered_raw_swagger, config

test_parameters = parsed_file_or_url('../test_data/protein.yaml')['test_parameters']


# TODO: clarify messaging.
def test_validate_and_call():
  try:
    bad_param_but_ok = defaultdict(list)
    good_param_not_ok = defaultdict(list)
    rs = parsed_file_or_url(config.swagger_path)
    paths = altered_raw_swagger(rs)['paths']
    for endpoint in paths:
        for verb in paths[endpoint]:
            assert verb in 'get post'
            validator = _validator(endpoint, verb)
            print(endpoint, verb)
            if endpoint in test_parameters:
                things = test_parameters[endpoint]
                for params in things['good']:
                    assert validator.is_valid(params)
                    print('   ok good valid', params)
                    response = call(endpoint, verb, params)
                    if not response.is_success:
                        good_param_not_ok[(endpoint, verb)].append(params)
                        raise ValidDataBadResponse(params)
                        """
                        {'rfActive': 'true'}   Returns a bad response.
                        {'rfActive': True}   Fails validation.
                        {'rfActive': True}   Returns a good response?
                        BUT This fucked.
                        The httpx (or some other Python lib) insists on using
                        True instead of 'true'.
                        But what the api actually wants is 'true', the json
                        version of True.
                        NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO 
                        NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO 
                        NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO 
                        The above returns a 500 response.
                        I have no idea if it is because of 'true' vs True.
                        But I think NOT.
                        Look at the test data.  It shows the response.
                        """
                    if response.is_success:
                        print('   ok good call')
                for params in things['bad']:
                    assert not validator.is_valid(params)
                    print('   ok bad NOT valid', params)
                    try:
                        response = call(endpoint, verb, params)
                    except (NonDictArgs, KeyError):
                        continue
                    if response.is_success:
                        bad_param_but_ok[(endpoint, verb)].append(params)
    bad_param_but_ok = dict(bad_param_but_ok)
    good_param_not_ok = dict(good_param_not_ok)
  finally:
    globals().update(locals())


def test_altered_raw_swagger():
    return
#     jdoc = raw_swagger(config.swagger_path)
#     jdoc = altered_raw_swagger(jdoc)
#     assert jdoc['paths']['/das/s4entry']['get']['parameters'] == []
#     assert jdoc['paths']['/']['get']['parameters'] == []





from datetime import datetime

from pyapix.tool import api_tools
from pyapix.tool.api_tools import (dynamic_validator, dynamic_call, SurpriseArgs)
from pyapix.tool.tools import (LocalValidationError, ValidDataBadResponse, )
from pyapix.client.info import local
from pyapix.tool.api_tools import endpoints_ands_verbs, Service
from pyapix.tool.tools import parsed_file_or_url


def local_validate(params):
    return


def altered_raw_swagger(jdoc):
    return jdoc

        
def head_func(endpoint, verb):
    return {'user-agent': 'python-httpx/0.27.2'}


class config:
    swagger_path = '~/osdu/service/unit-service/docs/v3/api_spec/unit_service_openapi_v3.json'
    api_base = 'https://yoohoo' 
    alt_swagger = altered_raw_swagger
    head_func = head_func
    validate = local_validate



name = 'Unit'
ends = endpoints_ands_verbs(parsed_file_or_url(config.swagger_path))
_validator = dynamic_validator(config)
call = dynamic_call(config)
service = Service(name, call, _validator, ends)


# aaaaaaaaarg!
# Seems impossible to delete this way.
# The only way to delete a buncha stuff then is one at a time.
#
# Solution...
# define service directly here.



# import copy
# #snapshot = copy.deepcopy(globals().items())
# snapshot = list(globals())
# to_delete = []
# #for (name, thing) in snapshot:
# for name in snapshot:
#     if (name not in ['call', '_validator', 'ends']) and not name.startswith('__'):
#         to_delete.append(name)
# #for thing in to_delete: del thing
# for name in globals():
#     if (name not in ['call', '_validator', 'ends']) and not name.startswith('__'):
#         try:
#             del globals()[name]
#         except RuntimeError:
#             print(name)
# 




# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from collections import defaultdict

import json
import jsonref

from pyapix.tool.tools import parsed_file_or_url
from pyapix.tool.api_tools import (NonDictArgs, ValidDataBadResponse,)
from pyapix.client.obis import _validator, call, config, altered_raw_swagger

test_parameters = parsed_file_or_url('../test_data/obis.yaml')['test_parameters']


def test_validate_and_call():
  try:
    bad_param_but_ok = defaultdict(list)
    good_param_not_ok = defaultdict(list)
    jdoc = parsed_file_or_url(config.swagger_path)  # TODO: pass flag for deref vs not.?
    jdoc = jsonref.loads(json.dumps(jdoc))
    paths = altered_raw_swagger(jdoc)['paths']
    for endpoint in paths:
        for verb in paths[endpoint]:
            print(endpoint, verb)
            validator = _validator(endpoint, verb)
            if endpoint in test_parameters:
                things = test_parameters[endpoint]
                for params in things['good']:
                    if not validator.is_valid(params):
                        validator.validate(params)

                    print('   ok good valid', params)
                    response = call(endpoint, verb, params)
                    if not response.is_success:
                        good_param_not_ok[(endpoint, verb)].append(params)
                        raise ValidDataBadResponse(params)
                    if response.is_success:
                        print('   ok good call')
                for params in things['bad']:
                    assert not validator.is_valid(params)
                    print('   ok bad NOT valid', params)
                    try:
                        response = call(endpoint, verb, params)
                    except (NonDictArgs, KeyError):
                        continue
#                        break
                    if response.is_success:
                        bad_param_but_ok[(endpoint, verb)].append(params)
  finally:
    bad_param_but_ok = dict(bad_param_but_ok)
    good_param_not_ok = dict(good_param_not_ok)
    globals().update(locals())





import doctest

import test_petstore
from pyapix.tool import working_with_postman
from pyapix.tool import do_postman_osdu
from pyapix.tool import do_postman_nws
from pyapix.osdu import crs_api
 

def test_it():
    dps = [do_postman_nws, do_postman_osdu]
    mods = [working_with_postman, crs_api] + dps

    for mod in mods:
        doctest.testmod(mod)

    for mod in dps:
        working_with_postman.check_do_item(mod.pm_files)
        mod.test_fetch_thing()

    working_with_postman.test_environment_update()

    crs_api.inspect_swagger()
    crs_api.inspect_postman()
    crs_api.test_crs_conversion()

#     test_petstore.test_seq()


test_it()



# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from apix.dyno import (dynamic_validator, dynamic_call,)

from info import local


class config:
    swagger_path = local.swagger.libre
    api_base = local.api_base.libre
    alt_swagger = lambda x:x
    head_func = lambda endpoint, verb: {'accept': 'application/json'}
    validate = lambda params: None


_validator = dynamic_validator(config)
call = dynamic_call(config)


# test
# ############################################################################
from collections import defaultdict
import json
import jsonref     # cross platform
from jsonschema import (validate, FormatChecker,)

from apix.tools import (parsed_file_or_url, ValidDataBadResponse)
from apix.dyno import NonDictArgs
from test_data.libre import test_parameters


def test_validate_and_call():
  try:
    bad_param_but_ok = defaultdict(list)
    good_param_not_ok = defaultdict(list)
    jdoc = parsed_file_or_url(config.swagger_path)
    jdoc = jsonref.loads(json.dumps(jdoc))
    paths = config.alt_swagger(jdoc)['paths']
    for endpoint in paths:
        for verb in paths[endpoint]:
            assert verb in 'get post'
            validator = _validator(endpoint, verb)
            print(endpoint, verb)
            if endpoint in test_parameters:
                things = test_parameters[endpoint]
                for params in things['good']:
                    if not validator.is_valid(params):
                        validator.validate(params)
                    print('   ok good valid', params)
                    response = call(endpoint, verb, params)
                    if not response.is_success:
                        # good_param_not_ok[(endpoint, verb)].append(params)
                        good_param_not_ok[(endpoint, verb)].append((params,
                                                                    response.text))
#                        raise ValidDataBadResponse(params)
                        # TODO: why the bad response !!!!!!!!!??????????
                        # TODO: why the bad response !!!!!!!!!??????????
                        # I want to sort this out because libre uses multiple
                        # verbs.
                        # TODO: why the bad response !!!!!!!!!??????????
                        # TODO: why the bad response !!!!!!!!!??????????
                    if response.is_success:
                        print('   ok good call')
                for params in things['bad']:
                    assert not validator.is_valid(params)
                    print('   ok bad NOT valid', params)
                    try:
                        response = call(endpoint, verb, params)
                    except NonDictArgs:
                        break
                    if response.is_success:
                        bad_param_but_ok[(endpoint, verb)].append(params)
  finally:
    bad_param_but_ok = dict(bad_param_but_ok)
    good_param_not_ok = dict(good_param_not_ok)
    globals().update(locals())


#if __name__ == '__main__': _validate_and_call()


# head = {'accept': 'application/json'}  # 
# api_file_path = local.swagger.libre
# api_base = local.api_base.libre
# 
# 
# def validate_and_call():
#   try:
#     rs = parsed_file_or_url(api_file_path)       # 
#     with httpx.Client(base_url=api_base) as client:   # 
#         verb_map = dict(get=client.get, post=client.post)
#         for ep in rs['paths']:
# #            print(ep)
#             ep_info = rs['paths'][ep]
#             print(ep, list(ep_info.keys()))
#             ep0 = ep
#             is_valid = validator_func(api_file_path, ep)
# #            print('     ', is_valid.schema)
#             if ep in test_parameters:
#                 things = test_parameters[ep]
#                 if ep0 != ep:
#                     print('   calling .............', ep)
#                 for params in things['good']:
#                     assert is_valid(params)
#                     print('   ok good valid', params)
#                     fetch = verb_map[list(ep_info.keys())[0]]
#                     r = fetch(ep, params=params, headers=head)
#                     assert r.status_code == 200
#                     # running locally
# 
#                     rj = r.json()
#                 for params in things['bad']:
#                     assert is_valid(params)  # TODO: fix
#                     assert is_valid(params)  # TODO: fix
#                     assert is_valid(params)  # TODO: fix
#                     print('   grrr bad but VALID', params)
#                     r = client.get(ep, params=params)
#                     assert r.status_code != 404    # Bad endpoint
#                     assert r.status_code in [400, 500]    # Bad Parameter
#   finally:
#     globals().update(locals())
# 
# 
#if __name__ == '__main__': validate_and_call()

# from jinja2 import Environment, select_autoescape 
# # TODO: remove in favor of newer thing.
# # endpoint_QUERY_params
# def insert_endpoint_params(endpoint, parameters):
#     if not '{' in endpoint:
#         return endpoint
#     env = Environment(autoescape=select_autoescape())
#     template = env.from_string(templatified(endpoint))
#     return template.render(**parameters)
# 
# 
# head = {'accept': 'application/json'}  # 
# api_file_path = local.swagger.libre
# api_base = local.api_base.libre
# 
# 
# 
# def schema_trans(schema_list):
#     d = {}
#     for thing in schema_list:
#         d['name'] = thing['schema'] if 'schema' in thing else {}
#     return {'properties': d}
# 
# 
# def validator_func(openapi_file, endpoint):
#   try:
#     """Return a function to validata parameters for `endpoint`.
#     """
#     rs = parsed_file_or_url(openapi_file)     # protein vs nws vs libre
#     with_refs = jsonref.loads(json.dumps(rs))
#     thing = with_refs['paths'][endpoint]
#     assert len(thing.keys()) == 1
#     verb = list(thing.keys())[0]
#     tv = thing[verb]
#     if 'parameters' in tv:
#         vinfo = tv['parameters']
#     else:
#         vinfo = {}
#     schema = schema_trans(vinfo)     # 
#     assert list(schema.keys()) == ['properties']
# #    print(endpoint, list(schema['properties'].keys()))
#     is_valid = lambda ob: Draft7Validator(schema, format_checker=FormatChecker()).is_valid(ob)
#     return is_valid
#   finally:
#     globals().update(locals())
#     is_valid.endpoint = endpoint
#     is_valid.schema = schema
# 
# 



# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from pyapix.tool.tools import (LocalValidationError,)
from pyapix.tool.api_tools import (dynamic_validator, dynamic_call,)
from .info import local

class ValidDataBadResponse(LocalValidationError): pass

class NonTruthy(LocalValidationError): pass

class InvalidAccessionId(LocalValidationError): pass


def local_validate(params):
    """Catch data problems missed by the schema.
    """
    if not params:
        raise NonTruthy(params)
    if params == {'accession': 'xxxxxxxx'}:
        raise InvalidAccessionId(params)


def altered_raw_swagger(jdoc):
    """Alter raw data to conform with local code assumptions.
    """
    patch = dict(parameters=[])
    jdoc['paths']['/das/s4entry']['get'].update(patch)
    jdoc['paths']['/']['get'].update(patch)
    return jdoc


class config:
    swagger_path = local.swagger.protein
    api_base = local.api_base.protein
    alt_swagger = altered_raw_swagger
    head_func = lambda endpoint, verb: {}
    validate = local_validate


_validator = dynamic_validator(config)
call = dynamic_call(config)






# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

# WoRMS: World Register of Marine Species

from pyapix.tool.api_tools import dynamic_validator, dynamic_call
from .info import local


class config:
    swagger_path = local.swagger.worms
    api_base = local.api_base.worms
    alt_swagger = lambda x: x 
    head_func = lambda endpoint, verb: {}
    validate = lambda params: None


_validator = dynamic_validator(config)
call = dynamic_call(config)



from datetime import datetime

from pyapix.tool import api_tools
from pyapix.tool.api_tools import (dynamic_validator, dynamic_call, SurpriseArgs)
from pyapix.tool.tools import (LocalValidationError, ValidDataBadResponse, )
from pyapix.client.info import local
from pyapix.tool.tools import parsed_file_or_url
from pyapix.tool.api_tools import endpoints_ands_verbs, Service


def local_validate(params):
    return


def altered_raw_swagger(jdoc):
    return jdoc

        
def head_func(endpoint, verb):
    return {'user-agent': 'python-httpx/0.27.2'}


class config:
    swagger_path = '~/osdu/service/crs-catalog-service/docs/api_spec/crs-catalog-openapi-v3.yaml'
    api_base = 'https://yoohoo' 
    alt_swagger = altered_raw_swagger
    head_func = head_func
    validate = local_validate


ends = endpoints_ands_verbs(parsed_file_or_url(config.swagger_path))
_validator = dynamic_validator(config)
call = dynamic_call(config)
service = Service('CRS Catalog', call, _validator, ends)




# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

import pytest
import httpx

from pyapix.tool import api_tools
from pyapix.tool.api_tools import (
    parameters_to_schema, 
    dynamic_call, 
    dynamic_validator,
    prep_func,
    BadEndpointOrVerb,
)
import pyapix.client
import pyapix.client.petstore


def test_dvalidator(): 
    print('test_dvalidator')
    # test using worms API
    class config:
        swagger_path = '~/local/worms/openapi.yaml'
        alt_swagger = lambda x: x 
        validate = lambda params: None

    dv = dynamic_validator(config)

    with pytest.raises(BadEndpointOrVerb):
        endpoint, verb = 'foo', 'get'
        dv(endpoint, verb)
    with pytest.raises(BadEndpointOrVerb):
        endpoint, verb = '/AphiaClassificationByAphiaID/{ID}', 'foo'
        dv(endpoint, verb)

    endpoint, verb = '/AphiaClassificationByAphiaID/{ID}', 'get'
    validator = dv(endpoint, verb)
    assert validator.is_valid({'ID': 127160})
    assert not validator.is_valid({'id': 127160})

    endpoint, verb = '/AphiaRecordsByName/{ScientificName}', 'get'
    validator = dv(endpoint, verb)
    assert validator.is_valid({'ScientificName': 'Solea solea'})
    assert not validator.is_valid({'taxonid': 0})

    # test using petstore API
    class config:
        swagger_path = '~/local/petstore/swagger.json'
        alt_swagger = lambda x: x 
        validate = lambda params: None
        api_base = 'petstore'

    dv = pyapix.client.petstore.dynamic_validator(config)

    endpoint, verb = '/pet/findByStatus', 'get'
    validator = dv(endpoint, verb)
    args = dict(status= ['sold', 'pending'])
    args = ['sold', 'pending']
    globals().update(locals())
    validator.validate(args)
    assert validator.is_valid(args)
    print('end test_dvalidator')


def param_func(pdict):
  try:
    d = {}
    for key in ['required', 'name', 'schema']:
        if key in pdict:
            d[key] = pdict[key]
    if 'schema' not in d:
        d['schema'] = {}
        for key in ['format', 'enum', 'type', 'items', 'default']:
            if key in pdict:
                d['schema'][key] = pdict[key]
    return d
  finally:
    globals().update(locals())


def altered_raw_swagger_petstore(jdoc):
    for endpoint in jdoc['paths']:
        for verb in jdoc['paths'][endpoint]:
            parameters = jdoc['paths'][endpoint][verb]['parameters']
            assert type(parameters) is list
            assert all(type(thing) is dict for thing in parameters)
            jdoc['paths'][endpoint][verb]['parameters'] = [param_func(d) for d in parameters]
    return jdoc


def test_prep_func(): 
  try:
    print('################################### start test_prep_func')
    # test using worms API
    class config:
        swagger_path = '~/local/worms/openapi.yaml'
        alt_swagger = lambda x: x 
        api_base = ''
        head_func = lambda *pos, **kw: {}

    pd = prep_func(config)

    endpoint, verb = '/AphiaClassificationByAphiaID/{ID}', 'get'
    args = {'ID': 127160}
    (u, v, request_params) = pd(endpoint, verb, args)
    assert u.endswith('/AphiaClassificationByAphiaID/127160')
#    assert request_params == {}
    # OK

    # test using petstore API
    class config:
        swagger_path = '~/local/petstore/swagger.json'
        alt_swagger = lambda x: x 
        validate = lambda params: None
        api_base = 'https://petstore.swagger.io/v2'
        head_func = lambda *pos, **kw: {}

    pd = pyapix.client.petstore.prep_func(config)

    uploadImage = False
    uploadImage = True
    endpoint, verb = '/pet/{petId}/uploadImage', 'post'
    print(endpoint, verb)
    if uploadImage:
        args = {'petId': 1234}
        print('  ', args)
        (u, v, request_params) = pd(endpoint, verb, args)
        assert u.endswith('/pet/1234/uploadImage')
        assert request_params == {}
        # OK
        # check if the request works...
        dc = dynamic_call(config)
        response = dc(endpoint, verb, args)
        assert response.status_code == 415
        assert response.reason_phrase == 'Unsupported Media Type'  # because we sent nothing.
        # OK
        print('  ', 'OK')

    if uploadImage:
        args = {'petId': 1234, 'additionalMetadata': 'aaaaaaa', 'file': 'foo.png'}
        print('  ', args)
        (u, v, request_params) = pd(endpoint, verb, args)
        assert u.endswith('/pet/1234/uploadImage')
        assert request_params == {'data': {'additionalMetadata': 'aaaaaaa'}, 'files': {'file': 'foo.png'}}

        # check if the request works...
        dc = dynamic_call(config)
        response = dc(endpoint, verb, args)
        assert response.status_code == 200
        assert response.json()['message'] == 'additionalMetadata: aaaaaaa\nFile uploaded to ./upload, 7 bytes'
        # OK
        print('  ', 'OK')


    biggie = False
    biggie = True
    if biggie:

        endpoint, verb = '/pet', 'post'
        print(endpoint, verb)
        args = {'id': 1234, 'category': {}, 'name': 'fluff', 'photoUrls': [], 'status': 'available', 'tags': []}
        print('  ', args)
        (u, v, request_params) = pd(endpoint, verb, args)
        assert request_params == {'json': args}
        # TODO: this one also seems fishy ???????
        assert list(request_params) == ['json']
        print(endpoint, verb)
        print('  args', args)
        print('  prep', request_params)
        print()
        # INVESTIGATE

        # checking if the request works...
        dc = dynamic_call(config)
        response = dc(endpoint, verb, args)
        assert response.status_code == 200
    #    assert response.json() == args
        post_response = response
        post_args = args

        # a GET request to verify the resource was created.
        endpoint, verb = '/pet/{petId}', 'get'
        args = {'petId': 1234}
        (u, v, request_params) = pd(endpoint, verb, args)
        response = dc(endpoint, verb, args)
        assert response.status_code == 200
        get_response = response

        assert post_response.json() == get_response.json()
        assert get_response.json()['name'] == 'fluff'
        print('   OK')
        # OK

        endpoint, verb = '/pet', 'put'
        args = {'id': 1234, 'name': 'buff'}
        (u, v, request_params) = pd(endpoint, verb, args)
        assert request_params == {'json': args}
        print(endpoint, verb)
        print('  args', args)
        print('  prep', request_params)
        print()
        # INVESTIGATE

        response = dc(endpoint, verb, args)
        assert response.status_code == 200
        assert response.json() == {'id': 1234, 'name': 'buff', 'photoUrls': [], 'tags': []}
        put_response = response
        print('   OK')
        # OK

        # GET request to verify the resource was altered.
        endpoint, verb = '/pet/{petId}', 'get'
        args = {'petId': 1234}
        (u, v, request_params) = pd(endpoint, verb, args)
        response = dc(endpoint, verb, args)
        assert response.status_code == 200
        get_response2 = response
        assert put_response.json() == get_response2.json()
        assert get_response2.json()['name'] == 'buff'
        print('   OK')
        # OK

        endpoint, verb = '/pet/{petId}', 'delete'
        args = {'petId': 1234, 'api_key': 'special-key'}
        (u, v, request_params) = pd(endpoint, verb, args)
        assert request_params == {'headers': {'api_key': 'special-key'}}

        # The resource is deleted, without sending the api_key.
        # BUT
        # This should be changed in prep_func.

        print(endpoint, verb)
        print('  args', args)
        print('  prep', request_params)
        request = httpx.Request(v, u, **request_params)
        dc = dynamic_call(config)
        response = dc(endpoint, verb, args)
        delete_response = response
        assert response.status_code == 200
        assert delete_response.json() == {'code': 200, 'type': 'unknown', 'message': '1234'}
        print('   OK')
        # OK

        # GET request to verify the resource was deleted.
        endpoint, verb = '/pet/{petId}', 'get'
        args = {'petId': 1234}
        (u, v, request_params) = pd(endpoint, verb, args)
        response = dc(endpoint, verb, args)
        assert response.status_code == 404
        assert response.reason_phrase == 'Not Found'
        get_response3 = response
        # OK


    endpoint, verb = '/pet/{petId}', 'delete'
    args = {'petId': 1234, 'api_key': 'special-key'}
    (u, v, request_params) = pd(endpoint, verb, args)


    endpoint, verb = '/pet/findByStatus', 'get'
    args = {'status': ['sold', 'pending']}

    (u, v, request_params) = pd(endpoint, verb, args)
    print(endpoint, verb)
    print('  args', args)
    print('  prep', request_params)
    print()

    request = httpx.Request(v, u, **request_params)
    assert str(request.url).endswith('/findByStatus?status=sold&status=pending')

    dc = dynamic_call(config)
    response = dc(endpoint, verb, args)
    rj = response.json()
    assert response.is_success
    assert len(rj) > 1
    # OK

    print('############################################# end test_prep_func')
  finally:
    globals().update(locals())


def test_dynamic_call():
    # TODO: ensure correct petstore config.
    class config:
        swagger_path = '~/local/petstore/swagger.json'
        alt_swagger = lambda x: x 
        validate = lambda params: None
        api_base = 'https://petstore.swagger.io/v2'
        head_func = lambda *pos, **kw: {}

    print('########################################### start test_dynamic_call')
    dc = dynamic_call(config)
    endpoint, verb = '/pet/findByStatus', 'get'
    args = dict(status= ['sold', 'pending'])
    response = dc(endpoint, verb, args)
    print('response', response)
    print('########################################### end test_dynamic_call')


def test_parameters_to_schema():
  try:
    """
    Petstore data
    Verifies function as implemented.  Which is OK, but ...
      could be better.
    """
    parameters = [{'description': 'Created user object',
  'in': 'body',
  'name': 'body',
  'required': True,
  'schema': {'properties': {'email': {'type': 'string'},
                            'firstName': {'type': 'string'},
                            'id': {'format': 'int64', 'type': 'integer'},
                            'lastName': {'type': 'string'},
                            'password': {'type': 'string'},
                            'phone': {'type': 'string'},
                            'userStatus': {'description': 'User Status',
                                           'format': 'int32',
                                           'type': 'integer'},
                            'username': {'type': 'string'}},
             'type': 'object',
             'xml': {'name': 'User'}}}]
    ev_info = dict(parameters=parameters)
    schema = pyapix.client.petstore.parameters_to_schema(ev_info)
#    assert sorted(list(schema)) == ['additionalProperties', 'properties', 'required', 'type']
    assert schema['properties'] == parameters[0]['schema']['properties']

    parameters = [{'name': 'status', 'in': 'query', 'description': 'Status values that need to be considered for filter', 'required': True, 'type': 'array', 'items': {'type': 'string', 'enum': ['available', 'pending', 'sold'], 'default': 'available'}, 'collectionFormat': 'multi'}]
    ev_info = dict(parameters=parameters)
    schema = parameters_to_schema(ev_info)
#    assert schema == {'name': 'status', 'in': 'query', 'description': 'Status values that need to be considered for filter', 'required': True, 'type': 'array', 'items': {'type': 'string', 'enum': ['available', 'pending', 'sold'], 'default': 'available'}, 'collectionFormat': 'multi'}
#    assert schema == parameters[0]

  finally:
    globals().update(locals())


test_parameters_to_schema()
test_dvalidator()
test_prep_func()
test_dynamic_call()




# Grab code from example/test and insert into documentation template.
# This way we know that the code appearing in documentation is correct.
#

import inspect
from jinja2 import Environment, select_autoescape 
from pyapix.examples import test_worms

def go():
    # TODO: 
    # Execute the test code.

    # Get test code.
    sc = inspect.getsource(test_worms.test_examples)
    test_code = '\n'.join(sc.split('\n')[1:])

    # Get the template.
    other = 'docs/using_template.md'
    with open(other) as fh:
        raw_other = fh.read()

    # Insert code into template.
    env = Environment(autoescape=select_autoescape())
    template = env.from_string(raw_other)
    stuffs = dict(worms_example=test_code)
    foo = template.render(**stuffs)
    bar = foo.replace('&#39;', '"')

    # Write documentation.
    with open('docs/using_it.md', 'w') as fh:
        fh.write(bar)

    # TODO: 
    # Generate using mkdocs.

go()





# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from datetime import datetime

try:
    from pyapix.tool.tools import LocalValidationError
    from pyapix.tool.api_tools import dynamic_validator, dynamic_call
    from .info import local
except:
    from .tools import LocalValidationError
    from .api_tools import dynamic_validator, dynamic_call
    from .info import local


class DateOrderError(LocalValidationError): pass


def local_validate(params):
    """Catch data problems missed by the schema.
    # eg start_date > end_date
    params = {
        'start': '2024-09-17T18:39:00+00:00', 
        'end':   '2024-09-18T18:39:00+00:00',
    }
    """
    fmt = '%Y-%m-%dT%H:%M:%S+00:00'
    if 'start' in params and 'end' in params:
        start = params['start']
        end = params['end']
        if datetime.strptime(start, fmt) > datetime.strptime(end, fmt): 
            raise DateOrderError(start, end)


def altered_raw_swagger(jdoc):
    """Alter raw data to conform with local code assumptions.
    This function takes a swagger doc as a json and returns json.
    """
    for endpoint in jdoc['paths']:
        epdoc = jdoc['paths'][endpoint]
        assert 'get' in epdoc
        assert 'parameters' in epdoc['get']
        if 'parameters' in epdoc:
            eprams = epdoc.pop('parameters')
            jdoc['paths'][endpoint]['get']['parameters'].extend(eprams)
    return jdoc

        
def head_func(endpoint, verb):
    """nws requires user-agent header.   Returns 403 otherwise.
    """
    return {'user-agent': 'python-httpx/0.27.2'}


class config:
    swagger_path = local.swagger.nws
    api_base = local.api_base.nws
    alt_swagger = altered_raw_swagger
    head_func = head_func
    validate = local_validate


_validator = dynamic_validator(config)
call = dynamic_call(config)

from pyapix.tool.tools import parsed_file_or_url
from pyapix.tool.api_tools import endpoints_and_verbs, Service
jdoc = parsed_file_or_url(config.swagger_path)
ends = endpoints_and_verbs(jdoc)
service = Service('NWS', call, _validator, ends)

# vd = _validator(endpoint, verb)
# vd.is_valid(parameters)
# vd.validate(parameters)
# call(endpoint, verb, parameters)
# The behavior of the two is inconsistent.
# How about this?
# caller = dynamic_call(config) ...
# call = caller(endpoint, verb)
# response = call(parameters)
# I guess that would be called a convenience function?




# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

# TODO: rename tools.py   ->  jtools.py    ?????

import time
import json
import os
from types import SimpleNamespace

import yaml
import httpx
from jsonschema import Draft7Validator


class LocalValidationError(Exception): pass

class ValidDataBadResponse(LocalValidationError): pass

class RemoteFileReadException(Exception): pass

def identity_func(x): return x


def extract_from_dict_list(list_of_dict, key):
    return {d['name']: d[key] for d in list_of_dict if key in d}


def retry_call(n=3, tfun=lambda i:i):
    def _retry(func):
        def __retry(url, verb, request_params):
            i = 0
            response = SimpleNamespace(is_success=False)
            while not response.is_success and i < n:
                not_started = False
                try:
                    response = func(url, verb, request_params)
                except (httpx.ReadTimeout, httpx.ConnectTimeout):
                    # TODO: capture exceptions?
                    pass
                i += 1
                time.sleep(tfun(i))
            return response
        return __retry
    return _retry


def raw_remote_file(at_path):
    heads = {'user-agent': 'python-httpx/0.27.2'}
    # TODO: header is required by NWS.
    request = httpx.Request('get', at_path, headers=heads)
    with httpx.Client() as client:
        try:
            response = client.send(request)  
        except httpx.ConnectError as exc:
            raise RemoteFileReadException(*exc.args)
    the_doc = response.text
    if response.status_code != 200:
        raise RemoteFileReadException(the_doc)
    return the_doc


def raw_local_file(at_path):
    with open(os.path.expanduser(at_path)) as fh:
        return fh.read()


def raw_file_or_url(at_path):
    match at_path[:4]:    # structural pattern matching requires python 3.10+
        case 'http':
            return raw_remote_file(at_path)
        case _:
            return raw_local_file(at_path)


def parsed_file_or_url(at_path):
    match at_path[-5:]: 
        case '.json':
            return json.loads(raw_file_or_url(at_path))
        case _:
            return yaml.safe_load(raw_file_or_url(at_path))


def dvalidator(local_validate): 
    def local_is_valid(params):
        try:
            local_validate(params)
            return True
        except LocalValidationError:
            return False

    class D7V:
        def __init__(self, *pos, **kw):
            self.v = Draft7Validator(*pos, **kw)
        def is_valid(self, thing):
            if not self.v.is_valid(thing):
                return False
            return local_is_valid(thing)
        def validate(self, thing):
            self.v.validate(thing)
            return local_validate(thing)

    return D7V


def list_of_dict_to_dict(key='key', value='value'):
    """
    >>> dlist = [
    ...     dict(key='foo', value=2),
    ...     dict(key='bar', value=9),
    ...     dict(key='bat', value=8, x=3),
    ... ]
    ... 
    >>> one_dict = list_of_dict_to_dict()
    >>> assert one_dict(dlist) == {'foo': 2, 'bar': 9, 'bat': 8}

    >>> dlist = [
    ...     dict(clave='foo', v=2),
    ...     dict(clave='bar', v=9),
    ...     dict(clave='bat', v=8, x=3),
    ... ]
    ...
    >>> one_dict = list_of_dict_to_dict('clave', 'v')
    >>> assert one_dict(dlist) == {'foo': 2, 'bar': 9, 'bat': 8}

    """
    def inner(list_of_dict):
        return {d[key]: d[value] for d in list_of_dict}
    return inner




# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from functools import lru_cache
from collections import defaultdict
import pytest

from pyapix.tool.tools import parsed_file_or_url
from pyapix.tool.api_tools import NonDictArgs
from pyapix.client.nws import call, _validator, altered_raw_swagger, config

# TODO: consider making call behave like _validator.
# The behavior of the two is currently maybe surprising / inconsistent.
# but the code is currently minimal.
# but the change will be a one-liner.

pfile = '../test_data/nws.yaml'
test_parameters = parsed_file_or_url(pfile)['test_parameters']
sfile = '../test_data/nws_sequence.yaml'
sdata = parsed_file_or_url(sfile)



# TODO: clarify messaging.
def test_validate_and_call():
  try:
    bad_param_but_ok = defaultdict(list)
    good_param_not_ok = defaultdict(list)
    jdoc = parsed_file_or_url(config.swagger_path)  # TODO: pass flag for deref vs not.?
    paths = altered_raw_swagger(jdoc)['paths']
    for endpoint in paths:
        for verb in paths[endpoint]:
            assert verb in 'get post'
            validator = _validator(endpoint, verb)
            print(endpoint, verb)
            if endpoint in test_parameters:
                things = test_parameters[endpoint]
                for params in things['good']:
                    if not validator.is_valid(params):
                        validator.validate(params)
                    print('   ok good valid', params)
                    response = call(endpoint, verb, params)
                    if not response.is_success:
                        good_param_not_ok[(endpoint, verb)].append(params)
                        raise ValidDataBadResponse(params)
                    if response.is_success:
                        print('   ok good call')
                if endpoint == '/stations/{stationId}/observations':
                    break
                break  # before bad
                for params in things['bad']:
                    assert not validator.is_valid(params)
                    print('   ok bad NOT valid', params)
                    try:
                        # TODO: re-extract prepped args.   ?????
                        # NO.
                        # Maybe.
                        # But first get accustomed to debugging as-is.
                        # Should have better visibility there.
                        response = call(endpoint, verb, params)
                    except NonDictArgs:
                        break
                    if response.is_success:
                        bad_param_but_ok[(endpoint, verb)].append(params)
  finally:
    bad_param_but_ok = dict(bad_param_but_ok)
    good_param_not_ok = dict(good_param_not_ok)
    globals().update(locals())


def demo_current_alerts(area='CO', zone='COZ040', event='Red Flag Warning'):
  try:
    denver_zone = 'COZ040'
    params = dict(zone=zone)
    params = dict(area=area, event=event)
    params = dict(area=area)
    endpoint, verb = '/alerts/active', 'get'
    response = call(endpoint, verb, params).json()
    feats = response['features']    # The interesting part
    return feats
  finally:
    cp = [a['properties'] for a in feats]
    for a in cp:
        az = a['affectedZones']
        ac = a['category']
        ae = a['event']
        ah = a['headline']
        ad = a['description']
        ax = a['areaDesc']
        ae = a['event']
        ae = a['event']
        print(ah)
        print('='*55)
        print(ad)
        print('='*55)
        print(' ' + ax.replace(';', '\n'))
        print()
        print()
    globals().update(locals())
#     return response['features']    # The interesting part
#     return response['type']        # 'FeatureCollection'
#     return response['title']       # string
#     return response['@context']    # list of stuff
#     return response['updated']     # timestamp


def demo_nws_series():
  try:
    """ Get a series of observations suitable for putting in a pandas DF,
    and then a jupyter notebook.
    """
    import pandas
    # Data
    endpoint = '/stations/{stationId}/observations'
    stationId = 'KRCM'   # OK
    stationId = 'CO100'   # OK
    params = {                                # OK
        'stationId': stationId, 
#         'start': '2024-11-13:59:00+00:00', 
#         'end': '2024-11-14:59:00+00:00', 
        'limit':   50,
    }
    # Seems it may require recent timestamps.
    validator = _validator(endpoint, 'get')
    assert validator.is_valid(params)

    response = call(endpoint, 'get', params)
    assert response.status_code == 200

    # Extract desired data from response.
    final = []
    feats = response.json()['features']
    assert feats
    for ft in feats: 
        pt = ft['properties']
        for key in [ '@id', '@type', 'elevation', 'station', 'rawMessage', 'icon', 'presentWeather', 'cloudLayers', 'textDescription', ]:
            pt.pop(key)
        for key in pt:
            if type(pt[key]) is dict:
                pt[key] = pt[key]['value']
        final.append(pt)

    # Convert to dataframe.
    df = pandas.DataFrame(final)
#    assert df.shape[0] > 10
    assert df.shape[1] == 15
    return df
  finally:
    globals().update(locals())


def test_current_alerts():
    # Because returning a value from a test generates a pytest warning.
    demo_current_alerts()


def test_nws_series():
    # Because returning a value from a test generates a pytest warning.
    demo_nws_series()


from pyapix.client.nws import call, _validator, altered_raw_swagger, config
from pyapix.client import nws
from pyapix.tool.peemish import run_seq, sequence_creator
from pyapix.tool import peemish
from pyapix.osdu.working_with_postman import Environment

sdata = parsed_file_or_url(sfile)

def test_seq():
    create_sequence = sequence_creator(nws.service)
    cs = create_sequence(sdata['nws_sequence'])

    env = Environment()     # Update the env with sdata
    env.collection['N'] = sdata['N']
    run_seq(cs, env)
    print(env.current)

    # exploration below
    response = peemish.response
    prj = peemish.response.json()
    feats = prj['features']
    for feat in feats:
        props = feat['properties']
        temp = feat['properties']['temperature']['value']
    globals().update(locals())
#     prf = peemish.response.json()['features']
#     for feat in prf:
#         fp = feat['properties']
#         # contains different zones; fireZone, forecastZone, etc.
#         forecast = fp['forecast']
#         county = fp['county']
#         fire = fp['fireWeatherZone']
#         observation = fp['@id']
# 
#     fts = set(f['properties']['@type'] for f in prf)
#     pros = peemish.response.json()['observationStations']
#     for station in pros: pass
#     stations = dict(
#         station=station,
#         forecast=forecast,
#         county=county,
#         fire=fire,
#         observation=observation
#     )
    globals().update(locals())





# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

from collections import defaultdict
import json

import jsonref
import jsonschema

from pyapix.tool.tools import parsed_file_or_url
from pyapix.tool.api_tools import NonDictArgs, SurpriseArgs
from pyapix.client import petstore as petstore_api
from pyapix.client.petstore import _validator, call, config, altered_raw_swagger
from pyapix.tool.peemish import run_seq, sequence_creator
from pyapix.tool import peemish

test_data_file = '../test_data/petstore_data.yaml'
pet_sequence_file = '../test_data/pet_sequence.yaml'
petstore_data = parsed_file_or_url(test_data_file)
sequence_data = parsed_file_or_url(pet_sequence_file)


def test_validate_and_call():
  try:
    test_parameters = petstore_data['test_parameters']
    bad_param_but_ok = defaultdict(list)
    good_param_not_ok = defaultdict(list)
    surprise_args = defaultdict(list)
    jdoc = parsed_file_or_url(config.swagger_path) #TODO:flag for deref vs not.?
    jdoc = jsonref.loads(json.dumps(jdoc))
    paths = altered_raw_swagger(jdoc)['paths']
    for endpoint in paths:
        for verb in paths[endpoint]:

            validator = _validator(endpoint, verb)
            print(endpoint, verb)
            if endpoint in test_parameters:
                things = test_parameters[endpoint]
                for params in things[verb]['good']:
                    if not validator.is_valid(params):
                        validator.validate(params)

                    print('   ok good valid', params)
                    try:
                        response = call(endpoint, verb, params)
                    except SurpriseArgs as exc:
                        surprise_args[(endpoint, verb)].append(params)
                        continue
#                    break  # after first params

                    if not response.is_success:
                        good_param_not_ok[(endpoint, verb)].append(params)
#                        raise ValidDataBadResponse(params)
                        continue
                    if response.is_success:
                        print('   ok good call')
#                break  # before bad ones
                for params in things[verb]['bad']:
                    assert not validator.is_valid(params)
                    print('   ok bad NOT valid', params)
                    try:
                        response = call(endpoint, verb, params)
                        if response.is_success:
                            bad_param_but_ok[(endpoint, verb)].append(params)
                    except (NonDictArgs, KeyError):
                        continue
#        break  # after first endpoint

  finally:
    bad_param_but_ok = dict(bad_param_but_ok)
    good_param_not_ok = dict(good_param_not_ok)
    globals().update(locals())

from pyapix.osdu.working_with_postman import Environment

def test_seq(name=None):
  try:
#    create_sequence = sequence_creator(petstore_api, sequence_data)
    create_sequence = sequence_creator(petstore_api)
    snames = [name] if name else ['pet_other_sequence', 'pet_crud_sequence']
    globals().update(locals())
    env = Environment()
#    env.collection['N'] = sequence_data['N']
    for sn in snames:
        seq = create_sequence(sequence_data[sn])
        run_seq(seq, env)
  finally:
    # debugging...
    try:
        pto = dict(
            endpoint = peemish.endpoint,
            verb = peemish.verb,
            post_test = peemish.post_test,
            args = peemish.args,
        )
        post_test = peemish.post_test
        response = peemish.response
        raw_code = peemish.raw_code
    except:
        pass

    globals().update(locals())



"""
Working with Postman files.
The focus is on OSDU preshipping.
"""

import os
import json
from functools import lru_cache
import typing
from typing import TypedDict
from typing import Required, NotRequired

from pyapix.tool.tools import parsed_file_or_url
from pyapix.tool.exploratory import pop_inputs, pop_key

items_seen = []
requests_seen = []
all_requests = []


# TODO: rename?   This is used in other two other files.
def check_do_item(pm_files):
  try:
    global all_requests
    all_requests = []
    for fpath in pm_files():
        with open(fpath) as fh: 
            jdoc = json.load(fh)
        print(fpath.split('/')[-1].split('.')[0])
        do_item(jdoc)
        print()
  finally:
    globals().update(locals())


# TODO: review
def do_item(thing, indent=0):
  try:
    is_item_or_request_but_not_both(thing)
    if not 'item' in thing:
        return do_request(thing, indent)
    name = thing['name'] if 'name' in thing else 'base'
    global items_seen
    items_seen.append(name)
    assert not is_request(thing)
    assert has_items(thing)
    items = thing['item']
    assert type(items) is list
    print('i', ' '*indent, name, len(items))
    for ithing in items:
        assert type(ithing) is dict
        if 'item' in ithing:
            do_item(ithing, indent+4)
        else:   # it is a request
            print(' '*(indent+4), ithing['name'])
            do_request(ithing, indent+4)
  finally:
    globals().update(locals())


# untidy
# OK.  Successfully decoded all headers in OSDU.
def do_headers(headers):
  try:
    return {h['key']: h['value'] for h in headers}

    standard_header_keys = ['key', 'type', 'value']
    standard_header_keys = ['key', 'value']

    assert type(headers) is list
    if not headers:
        return headers
    assert len(headers) < 14
    for header in headers:
        assert all(k in header for k in standard_header_keys)
#        hkeys = sorted(list(header.keys()))
    return [sorted(list(h.keys())) for h in headers]
    return headers
  finally:
    globals().update(locals())


# untidy
def do_request(thing, indent):
  try:
    assert is_request(thing)
    global all_requests
    all_requests.append(thing)
    assert not has_items(thing)
    global requests_seen
    requests_seen.append(thing['name'])
    request = thing['request']
    assert type(request) is dict
    for word in ['method', 'header', 'url']:
        assert word in list(request)
    other_words = ['auth', 'body', 'description']  # may be in list(request)

    url_raw = request['url']['raw']
    print(' '*(indent+2), url_raw)

    dh = do_headers(request['header'])
#    print(' '*(indent+2), dh)
#    print(' '*(indent+2), len(dh))

    if 'body' in request:
        bm = request['body']['mode']
        bd = decode_body(request['body'])
    else:
        bm = bd = 'NO body'
#    print(' '*(indent+2), bm)
#    print(' '*(indent+2), bd)
    if 'auth' in request:
        ra = request['auth']
  finally:
    globals().update(locals())


def has_items(thing):
    return 'item' in thing

def is_request(thing):
    return 'request' in thing

verified_mutually_exclusive = []
def is_item_or_request_but_not_both(thing):
    assert is_request(thing) or has_items(thing)
    assert not (is_request(thing) and has_items(thing))
    global verified_mutually_exclusive
    
    name = thing['name'] if 'name' in thing else 'base'
    verified_mutually_exclusive.append(name)
    # TODO: this is a job for a closure.
    # TODO: appears to be capturing `items` but not `requests`.
    # pprint(set(verified_mutually_exclusive))


# TODO: if any of the url/path/query parsing stuff is needed, it is this.
def fix_colon_prefix(path):
  try:
    """
    Accomodate a Postman quirk.   Or is it an OSDU quirk?
    >>> path = '/foo/:bar/bat/:ratHatCat'
    >>> assert fix_colon_prefix(path) == '/foo/{{bar}}/bat/{{rat_hat_cat}}'
    """
    if ':' not in path:
        return path
    words = path.split('/')
    for (i, word) in enumerate(words):
        if word.startswith(':'):
            new = []
            for char in word[1:]:
                if char.isupper():
                    new.append('_')
                    new.append(char.lower())
                else:
                    new.append(char)
#            words[i] = '{{' + ''.join(new) + '}}'
            words[i] = '{' + ''.join(new) + '}'
    return '/'.join(words)
  finally:
    globals().update(locals())


"""
enames = ['Core Services', 'Entitlements']   # good output
# show_contents(pmjdoc, *enames)   # good output

>>> ebnames = ['Core Services', 'CRS Catalog', 'Entitlements']
TODO:  weird things happen with this input.  fix.
>>> show_contents(pmjdoc, *ebnames)
Core Services
    CRS Catalog
        Entitlements
            i V3
            r Health Check
>>> show_contents(pmjdoc, 'Core Services', 'CRS Catalog')
Core Services
    CRS Catalog
        i V3
        r Health Check
"""
# TODO: mv to where fetch_thing is.
# TODO: change name to show_pm_contents
# or pm_show_contents
# or pm.show_contents
def show_contents(pmjdoc, *names):
    """
    show_contents(pmjdoc)
    show_contents(pmjdoc, 'Core Services')
    show_contents(pmjdoc, 'Core Services', 'Entitlements')
    show_contents(pmjdoc, 'Core Services', 'CRS Catalog', 'V3')
    """
    pm_item = fetch_thing(pmjdoc, *names)
    space = ' '
    i = 0
    indent = space * i
    for name in names:
        print(f'{indent}{name}')
        i += 4
        indent = space * i
    if 'item' in pm_item:
        for dct in pm_item['item']:
            t = 'r' if 'request' in dct else 'i' 
            print(f"{indent}{t} {dct['name']}")


# solid below here.
# ######################################################################## #


# solid
def insert_params(template, parameters):
    """
    >>> url = '{{base_url}}/api/search/v2/query'
    >>> ps = dict(base_url='xxxxxxxx')
    >>> x = insert_params(url, ps)
    >>> assert x == 'xxxxxxxx/api/search/v2/query'

    >>> template = 'xyz {{abc}} wvp'
    >>> abc = 'xxxxxxxx'
    >>> x = insert_params(template, locals())
    >>> assert x == 'xyz xxxxxxxx wvp'
    """
    from jinja2 import select_autoescape 
    from jinja2 import Environment as j2Environment
    env = j2Environment(autoescape=select_autoescape())
    return env.from_string(template).render(**parameters)


# solid
class Environment:
    """A hierarchy of environments, ala Postman.
    >>> environment = Environment()
    >>> assert 'k' not in environment.general
    >>> assert 'k' not in environment.current
    >>> environment.request['k'] = 'v'
    >>> assert environment.current['k'] == 'v'
    >>> assert 'k' not in environment.general
    >>> environment.reset()
    >>> assert environment.current == {}
    >>> assert environment.request == {}
    """
    def __init__(self):
        self._current = {}
        self.general = {}
        self.collection = {}
        self.sequence = {}
        self.request = {}

    def update(self):
        for source in [self.general, self.collection, self.sequence, self.request]:
            self._current.update(source)
#        globals().update(self._current)    # ?

    @property
    def current(self):
        self.update()
        return self._current

    def reset(self):
        self.general = {}
        self.collection = {}
        self.sequence = {}
        self.request = {}
        self._current = {}


def test_environment():
    environment = Environment()
    assert environment.current == {}
    things = [
        (environment.general, None),
        (environment.collection, 2),
        (environment.sequence, 22),
        (environment.request, 222),
    ]
    for (thing, value) in things:
        thing['foo'] = value
    assert 'foo' in environment.current
    assert environment.current['foo'] == environment.request['foo']
    for (thing, value) in things:
        assert thing['foo'] == value
    environment.reset()
    assert environment.current == {}


def read_it():
    return environment.current['a']

def write_it():
    environment.request['a'] = 4

def setup_environment():
    environment = Environment()
    environment.collection['a'] = 1
    environment.request['a'] = 2
    globals().update(locals())

def test_environment2():
    """Demonstrate how to read/write a global environment.
    """
    setup_environment()
    assert read_it() == 2
    environment.request.pop('a')
    assert read_it() == 1
    environment.request['a'] = 3
    assert read_it() == 3
    write_it()
    assert read_it() == 4


# solid
# TODO: maybe call it fetch_postman_thing
def fetch_thing(jdoc, *names):
    """
    >>> innermost = [
    ...     dict(name='bat', t=1),
    ...     dict(name='x', t=2),
    ...     dict(name='y'),
    ... ]
    ... 
    >>> mid = [
    ...     dict(name='bar', item=innermost),
    ...     dict(name='c', t=3, item=[]),
    ...     dict(name='d'),
    ... ]
    ... 
    >>> outermost = dict(item=[
    ...     dict(name='foo', item=mid),
    ...     dict(name='a', item=[]),
    ...     dict(name='b'),
    ... ])
    ... 

    >>> assert fetch_thing(outermost) == outermost
    >>> assert fetch_thing(outermost, 'foo') == dict(name='foo', item=mid)
    >>> assert fetch_thing(outermost, 'foo', 'bar') == dict(name='bar', item=innermost)
    >>> assert fetch_thing(outermost, 'foo', 'bar', 'bat') == {'name': 'bat', 't': 1}
    """
    sub = jdoc
    for name in names:
        for thing in sub['item']:
            if thing['name'] == name:
                sub = thing
                break     # the first thing with that name
    return sub


# solid
# OK.  Successfully decoded all bodies in OSDU.
def decode_body(body):
    bm = body['mode']
    assert bm in ['raw', 'urlencoded', 'file']
    br = body[bm]
    if (type(br) is not str) or (not br):
        return br
    return json.loads(br)


# OK
# Now we can
# 1. Recursively iterate over all things.
# 2. Fetch arbitrary, deeply nested things.
# TODO: 
# - run individual request.
# - run a sequence of requests.
# - cleanup the jdoc by removing empty things.
# - add and subtract things.


# Questionable usefulness below
# ########################################################################## #

# used once.
def write_data():
    # TODO: write petstore to yaml.
    # !!!!!!!!!! WARNING !!!!!!!!!!
    # Use great caution when writing yaml because it can come out quite garbled.
    # All info will be there but with references not very readable.
#    from pyapix.test_data import petstore
    import yaml
    fname = 'petstore_dataX.yaml'
    data = petstore.__dict__
    for key in dubs:
        data.pop(key)
    with open (fname, 'w') as fh:
        yaml.dump(data, fh)

# not used
@lru_cache
def postman_schema():
    # all postman schemas == v1.0.0  v2.0.0  v2.1.0
    # OSDU Preshipping postman files have this...
    'https://schema.getpostman.com/json/collection/v2.1.0/collection.json'
    # for schema.  But that link is   301 Permanently moved.
    postman_schema = 'https://schema.postman.com/collection/json/v2.1.0/draft-07/collection.json'
    return parsed_file_or_url(postman_schema)


# TODO: is this needed?
def decode_url(url):
    """For working with Postman.
    But should be much more general.
    """
    if not '?' in url:
        return (fix_colon_prefix(url), '')
    assert url.count('?') == 1
    front, end = url.split('?')
    parts = end.split('&')
    assert  all(len(x.split('='))==2 for x in parts)
    query_params = dict(x.split('=') for x in parts)
    front = fix_colon_prefix(front)
    return (front, query_params)


def test_decode_url():
  try:
    urls = """
/crs/catalog/v3/coordinate-reference-system?dataId=Geographic2D:EPSG::4158&recordId=osdu:reference-data--CoordinateReferenceSystem:Geographic2D:EPSG::4158
/register/v1/action/:id
/register/v1/action:retrieve
/register/v1/subscription/:id/secret
/unit/v3/unit/unitsystem?unitSystemName=English&ancestry=Length&offset=0&limit=100
/unit/v3/unit/measurement?ancestry=1
/unit/v3/conversion/abcd?namespaces=Energistics_UoM&fromSymbol=ppk&toSymbol=ppm
/unit/v3/conversion/abcd
/legal/v1/legaltags:query?valid=true
/entitlements/v2/groups/:groupEmail/members/:memberEmail
/entitlements/v2/members/:memberEmail/groups?type=DATA
    """.split()
    for url in urls:
        front, query_params = decode_url(url)
        print(url)
        print(front)
        print(query_params)
        print()

    # One-time dev stuff below.....
    eswagger = '~/osdu/service/entitlements/docs/api/entitlements_openapi.yaml'
    ejson = parsed_file_or_url(eswagger)
    eps = list(ejson['paths'])
    # AHA!    brainwave!!! 
    # The OSDU openapi files violate the standard thus...
    # /groups/:groupEmail/members/:memberEmail
    # should be
    # /groups/{group_email}/members/{member_email}
    # I guess they think they know better.

    lswagger = '~/osdu/service/legal/docs/api/legal_openapi.yaml'
    ljson = parsed_file_or_url(lswagger)
    lps = list(ljson['paths'])

    sp = '~/osdu/service/register/docs/api/register_openapi.yaml'
    js = parsed_file_or_url(sp)
    rps = list(js['paths'])

  finally:
    globals().update(locals())


def exp_with_postman_schema():
    pm_schema = parsed_file_or_url('~/local/postman/2.1.0.json')
    request_schema = pm_schema['definitions']['request']
    pm_defs = pm_schema['definitions']
    ['$schema', '$id', 'title', 'description', 'oneOf']
    ['url', 'auth', 'proxy', 'certificate', 'method', 'description', 'header', 'body']

    script_schema = pm_schema['definitions']['script']
    ['$schema', '$id', 'title', 'type', 'description', 'properties']
    ['id', 'type', 'exec', 'src', 'name']


def exp_with_TypedDict():
    Point2D = TypedDict('Point2D', {'in': int, 'x-y': int})
    # TODO: NOTE
    # This is interesting.  Keys that are keywords or contain '-'.
    # Could be useful when we want to use keys like 
    # /zones/:type/:zoneId
    Point2D = TypedDict('Point2D', {'/zones/:type/:zoneId': int, 'x-y': int})

    p2 = Point2D( {'z': 3, 'label': 'bad'})
    p2 = Point2D(t= 3, label= 'bad')


class Request(TypedDict, total=True):
    header: Required[dict]
    body: typing.Dict  = None
    description: str  = None
    method: str     # enum
    url: str     # matching a regex

class MyParameters(TypedDict, total=True):
    header: NotRequired[dict]
    body: NotRequired[dict]
    query: NotRequired[dict]
    args: NotRequired[dict] = {}
 
class MyRequest(TypedDict, total=True):
    endpoint: str     # matching a regex
    method: str     # enum
    parameters: MyParameters = None
    post_test:  typing.Callable   = lambda _:None
    # TODO?: auth: dict  # or such?

# MyRequest and MyParameters can be used together to create a Request object.
# That will be the mapping between my stuff and Postman schema.

mr = MyRequest()
mr = MyRequest(ep=1)
mr = MyRequest(endpoint=1, method='have', parameters={})

rt = Request()
rt = Request(x=2)
rt = Request(header=2, url='u', method='m')


#    Newly moved to this file.......


# TODO: mv to do_postman_osdu  DONE
def is_version(word):
    """
    >>> assert is_version('v3') is True
    >>> assert is_version('v222') is True
    >>> assert is_version('V222') is True
    >>> assert is_version('vx2') is False
    """
    if word[0].lower() == 'v' and word[1:].isdigit():
        return True
    return False


# TODO: mv to do_postman_osdu?    YES  DONE
def is_bad_schema(schema):
    if schema == { 'required': [], 'properties': {},
     'additionalProperties': False, 'type': 'object'}:
        return True
    if schema['properties'] == {} and schema['additionalProperties'] == False:
        return True
    return False


# TODO: mv with func below  DONE
def make_endpoint(url, service):
  try:
    eps = [e for (e,v) in service.ends]   # all endpoints for swagger
    up = url['path']
    endpoint = url['raw'].replace(url['host'][0], '')
    for i, word in enumerate(up):
        if is_version(word):
            svc = up[:i+1]
            endpoint = '/' + '/'.join(up[i+1:])
            # check here to see if swagger endpoint contains version or not.
            # Include the version if it is in the swagger file.
            if endpoint not in eps:
                for e in eps:
                    if e.endswith(endpoint):
                        endpoint = e
    return fix_colon_prefix(endpoint)
  finally:
    globals().update(locals())





# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

# Interactive repl
# #########################################################################

from api_libre import *
import cmd, sys
from turtle import *
import readline
import os
import httpx

# TODO: try cmd2 in place of cmd.
#import cmd2
#import gnureadline

histfile = os.path.expanduser('~/.trans_history')
histfile_size = 1000000


def translate(params):
    endpoint = '/translate'
    api_base = local.api_base.libre
    with httpx.Client(base_url=api_base) as client:
        r = client.post(endpoint, params=params)
        assert r.status_code == 200
    return r.json()


def es2ingles(phrase):
    params = {'q': phrase, 'source': 'es', 'target': 'en'}
    endpoint = '/translate'
    api_base = local.api_base.libre
    with httpx.Client(base_url=api_base) as client:
        r = client.post(endpoint, params=params)
    return r.json()['translatedText']

def en2spanish(phrase):
    params = {'q': phrase, 'source': 'en', 'target': 'es'}
    endpoint = '/translate'
    api_base = local.api_base.libre
    with httpx.Client(base_url=api_base) as client:
        r = client.post(endpoint, params=params)
    return r.json()['translatedText']


class TurtleShell(cmd.Cmd):
    intro = 'Welcome to the trans shell.   Type help or ? to list commands.\n'
    prompt = '(trans) '
    file = None
    mode = 'es'
    mode_map = dict(es=es2ingles, en=en2spanish)

    
# TODO: how to save cmd.Cmd history separate from ordinary python history.
#     # Saves regular python in addition to the cmd loop.
#     def preloop(self):
#         if readline and os.path.exists(histfile):
#             readline.read_history_file(histfile)
#  
#     def postloop(self):
#         if readline:
#             readline.set_history_length(histfile_size)
#             readline.write_history_file(histfile)


    # ----- commands -----
    def do_mode(self, arg):
        if arg:
            self.mode = arg
        msg = f'the current mode is : {self.mode}'
        print(msg)

    def default(self, arg):
        func = self.mode_map[self.mode]
        print(func(arg))

    def do_exit(self,*args):
        return True


if __name__ == '__main__': 
    TurtleShell().cmdloop()


# ['/detect', '/frontend/settings', '/languages', '/suggest', '/translate', '/translate_file'])

# def get_component_schemas_libre():    #  # TODO: rename func
#     rs = parsed_file_or_url(local.swagger.libre)
#     with_refs = jsonref.loads(json.dumps(rs))
#     return with_refs['definitions']

def languages():
    endpoint = '/languages'
    api_base = local.api_base.libre
    with httpx.Client(base_url=api_base) as client:
        r = client.get(endpoint, params=params)
        assert r.status_code == 200
    return r.json()





# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.


detect = {
    'good': [
        { 'q': 'What language is this?', },
        { 'q': 'Que idioma es?', },
    ],
    'bad': [],
}

translate = {
    'good': [
        { 'q': 'What language is this?', 'source': 'en', 'target': 'es', },
    ],
    'bad': [],
}

languages = {
    'good': [],
    'bad': [],
}

test_parameters = {
    '/detect': detect ,
    '/languages': languages ,
    '/translate': translate ,
} 




# Copyright (c) 2024-2025 Cary Miller
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.


class local:
    class swagger:
        petstore = '~/local/petstore/swagger.json'
        nws = '~/local/nws/openapi.json'
        protein = '~/local/ebi/protein_openapi.json'
        libre =  '~/local/libretranslate/openapi.json'
        obis =  '~/local/obis/obis_v3.yml'
        worms =  '~/local/worms/openapi.yaml'
    class api_base:
        petstore = 'https://petstore.swagger.io/v2'
        nws = 'https://api.weather.gov'
        protein = 'https://www.ebi.ac.uk/proteins/api'
        libre = 'http://localhost:5000'
        obis = 'https://api.obis.org/v3'
        worms = 'https://www.marinespecies.org/rest'

class remote:
    class swagger:
        petstore = 'https://petstore.swagger.io/v2/swagger.json'
        nws = 'https://api.weather.gov/openapi.json'
        interpro = 'https://www.ebi.ac.uk/interpro/api/static_files/interpro7-swagger.yml'
#        protein = 'cannot find it'
#        libre = 'https://libretranslate.com/docs/spec'
#        This is it.  Loads in browser w/o incident but not programatically.
        obis = 'https://api.obis.org/obis_v3.yml'
        worms = 'https://www.marinespecies.org/rest/api-docs/openapi.yaml'
        

class common:
    class headers:
        class content_type:
            json = {'Content-Type': 'application/json'}
            form_data = {'Content-Type': 'form-data'}
        class accept:
            json = {'Accept': 'application/json'}





# https://nox.thea.codes/en/stable/tutorial.html
# ^ The nox docs ^
# Here is a noxfile. >
# Pretty much follows the tutorial.

import os
import nox


def pyfiles(dirname):
    cwd = os.getcwd()
    fnames = [fn for fn in os.listdir('apis') if fn.endswith('.py')]
    return [os.path.join(cwd, dirname, fn) for fn in fnames]


@nox.session
def flake8(session):
    session.install('flake8')
    session.run('flake8', *pyfiles('apis'), *pyfiles('examples'))


# TODO: use ruff instead of flake8 + black + isort + ...
# It is made by the same group that made `uv`.
# Looks to be very nice.
@nox.session
def lint(session):
    session.install('ruff')
    cwd = os.getcwd()
    dir_names = ['apis', 'examples']
    dir_paths = [os.path.join(cwd, dn) for dn in dirs_names]
    session.run('ruff', 'check', *dir_paths)
# https://github.com/astral-sh/ruff
# TODO: NOTE ruff has a buncha configs to go in the pyproject.toml file.

@nox.session
def tests(session):
    session.install("-r", "requirements.txt")
    session.install(".")     # install the apis/ and examples/ directories.
    session.run("pytest")


@nox.session(tags=["fail", "fluff"])
@nox.parametrize("django", ["1.9", "2.0"])
def peps(session, django):
    print('yoohoo peps....... django ==', django)
#    session.install_and_run_script("peps.py")   # it fails.


#@nox.session(python=["3.11", "3.12"])
@nox.session(tags=["fluff"])
def foo(session):
    session.install("flake8")
    print('yoohoo foo')
#    session.run("flake8", env={"FLASK_DEBUG": "1"})  # bar never runs
    session.notify("bar", posargs=[1, 'a'])


@nox.session
def bar(session):
    session.install("flake8")
    print('yoohoo bar', *session.posargs)





"""
Use a NWS postman file.
# Below url is exportable & downloadable to json if logged in to postman.
url = 'https://www.postman.com/api-evangelist/national-oceanic-and-atmospheric-administration-noaa/documentation/9eu7ygi/weather-gov'
BUT
The thing above has NO parameters!
Fooeey!

# But this has parameters?
https://github.com/aisabel/PetStore-Postman
Looks like it does.
It has collection AND environment.
But is PUNY.

btw.  Postman is acting like assholes.  People are switching to Bruno.
https://www.usebruno.com
enshitification
Postman requires storing credentials in the cloud, apparently.
Because Postman now requires a cloud account.

"""
import os
from pyapix.tool.tools import parsed_file_or_url
from pyapix.tool import working_with_postman as wp
from pyapix.tool.working_with_postman import (
    is_request, fetch_thing,
)

def pm_files():
    files = ['~/local/Weather.gov.postman_collection.json']
    return [os.path.expanduser(f) for f in files]


def test_fetch_thing():
  try:
    jdoc = parsed_file_or_url(pm_files()[0])
    names = [
        'zones',
        '{type}',
        '{zone Id}',
        '/zones/:type/:zoneId/forecast',
        ]
    assert fetch_thing(jdoc) == jdoc
    zi = fetch_thing(jdoc, *names[:-1])
    zr = fetch_thing(jdoc, *names)
    assert str(zr) in str(zi)
    zrr = zr['request']
    assert list(zi) == ['name', 'item']
    assert list(zr) == ['name', 'request']
    assert list(zrr) == ['method', 'header', 'url', 'description']
    assert is_request(zr)
    assert not is_request(zi)
  finally:
    globals().update(locals())




from types import SimpleNamespace
import typing
import time

from pyapix.tool.tools import parsed_file_or_url
from pyapix.osdu.working_with_postman import insert_params


count = 0  # for assigning sequential ID to service_request.

class PostTestFailure(Exception): pass


class Sequence:
    """A sequence of calls to one or more API services.
    Leaving `call` in the service_request instead of Sequence is a good move.  
    Makes calling multiple services trivial.
    """
    auth = None

    def __init__(self, rseq):
        self.rseq = rseq   # list of requests

    def run_seq(self, env):
        for req in self.rseq:
            req.run(1, env)    # passing 1 for auto-retry.

    def show_names(self):
        for req in self.rseq:
            msg = f'{req.name:<22} {req.endpoint} {req.verb}'
            print(f'{msg:<55} {req.tested}')


def request_for_service(client):
    # TODO: Much of auth goes here.
    def arequest(name='', endpoint='', verb='', args=(), post_test=lambda _:None):
      try:
        # Quick way to ensure a dict with only specific keys.
        global count
        count += 1
        secret_id = count
        tested = 'untested'
        self = SimpleNamespace(locals())

        def run(i, environment):
          try:
            print(f'=========== {i} running request... {name}')
            # TODO: optional validation here.
            # TODO: insert to args from environment here. DONE
            for (nom, value) in args.items():
                if (nom in environment.current) and all(c in value for c in '{}'):
                    ip = insert_params(value, environment.current)
                    if ip != value:
                        args[nom] = ip
            
            response = client.call(endpoint, verb, args)
            fna = response
            nonlocal self
            try:
                # TODO: this retry logic is ugly.  FIX.
#                f = lambda response, environment: post_test(response)
#                self.tested = f(response, environment)
                try:
                    post_test.requires
                except AttributeError:
                    pass
                self.tested = post_test(response, environment)
                # TODO: environment
                # is relevant here.
            except AssertionError as exc:   # sleep before retry
                if i < 7:
                    time.sleep(5*i)
                    run(i+1)
                else:
                    raise PostTestFailure(exc)
            return tested
          finally:
            globals().update(locals())

        self.run = run
        return self
      finally:
        globals().update(locals())
    return arequest


def sequence_creator(client):     
    # TODO: The whole thing is petstore-centric.
    # SOLUTION:  WORMS+OBIS+ProteinDB
    service_request = request_for_service(client)
    # TODO: problem.
    # This limits a sequence to a single service.
    # Which is maybe not such a problem.
    # Multi-service sequences can be made by adding multiple single-service
    # sequences.
    def create_sequence(sequence):
      try:
        out_seq = []
        i = 0
        for dct in sequence:
            i += 1
            requires = dct['post_test']['requires']
            raw_code = dct['post_test']['code']
#            for key in requires:
#                line = f'{key} = {dct[key]}\n'    # NO
#                 line = key + f'= {key}\n'
#                 # TODO: key needs to be substituted from Environment
#                 raw_code = line + raw_code
#            post_test.requires = {key:dct[key] for key in requires}
            exec(raw_code, locals=dct) # eek!
            # TODO: NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO
            # TODO: NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO
            # Rather than exec and keep the function object, simply execute in
            # the presence of an Environment.
            # ?????????????????????
            # TODO: NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO
            # TODO: NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO NO
            pr = service_request(**dct) 
            out_seq.append(pr)
        return Sequence(out_seq)
      finally:
        globals().update(locals())
    return create_sequence


def run_seq(seq, env):
    seq.show_names()
    seq.run_seq(env)
    seq.show_names()
    print('\n'.join(['*'*55]*4))
    print()


